from typing import Callable, Any

from kallikrein.expectation import LiftExpectationResult, ExpectationResult
from kallikrein import Expectation

from amino import do, Do

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.util import nvimio_repeat_timeout
from ribosome.nvim.io.api import N
from ribosome.nvim.io.data import NError


def failure_result(expectation: Callable[..., NvimIO[Expectation]], *a: Any, **kw: Any) -> NvimIO[ExpectationResult]:
    def failure_result(result: NError[Expectation]) -> NvimIO[ExpectationResult]:
        return expectation(*a, **kw)
    return failure_result


@do(NvimIO[Expectation])
def eval_expectation(expectation: Callable[..., NvimIO[Expectation]], *a: Any, **kw: Any) -> Do:
    io = yield expectation(*a, **kw)
    yield N.from_io(io.evaluate)


@do(NvimIO[Expectation])
def await_k(
        expectation: Callable[..., NvimIO[Expectation]],
        *a: Any,
        timeout: int=1,
        interval: float=.25,
        **kw: Any,
) -> Do:
    yield N.recover_error(
        nvimio_repeat_timeout(
            lambda: eval_expectation(expectation, *a, **kw),
            lambda a: a.success,
            f'expectation not satisfied within {timeout} seconds',
            timeout,
            interval,
        ).map(LiftExpectationResult),
        failure_result(expectation, *a, **kw),
    )


__all__ = ('await_k',)



import abc
from uuid import uuid4, UUID
from typing import Tuple

from amino import Dat, Map, Boolean, ADT, List, Maybe, Lists
from amino.boolean import false

from ribosome.compute.program import Program


class MapMode(ADT['MapMode']):

    @abc.abstractproperty
    def mnemonic(self) -> str:
        ...


class mapmode:

    class Normal(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'n'

    class Operator(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'o'

    class Visual(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'x'

    class Select(MapMode):

        @property
        def mnemonic(self) -> str:
            return 's'

    class VisualSelect(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'v'

    class Insert(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'i'

    class CommandLine(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'c'

    class Terminal(MapMode):

        @property
        def mnemonic(self) -> str:
            return 't'

    class Language(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'l'


class Mapping(Dat['Mapping']):

    @staticmethod
    def cons(keys: str, buffer: Boolean=false, modes: List[MapMode]=None, uuid: UUID=None) -> 'Mapping':
        modes1 = List(mapmode.Normal()) if modes is None else modes
        return Mapping(keys, Boolean(buffer), modes1, uuid or uuid4())

    def __init__(self, keys: str, buffer: Boolean, modes: List[MapMode], uuid: UUID) -> None:
        self.keys = keys
        self.buffer = buffer
        self.modes = modes
        self.uuid = uuid


class Mappings(Dat['Mappings']):

    @staticmethod
    def cons(*mappings: Tuple[Mapping, Program]) -> 'Mappings':
        return Mappings(Map(Lists.wrap(mappings).map2(lambda m, p: (str(m.uuid), (m, p)))))

    def __init__(self, mappings: Map[str, Tuple[Mapping, Program]]) -> None:
        self.mappings = mappings

    def lift(self, mapping: Mapping) -> Maybe[Program]:
        return self.mappings.lift(str(mapping.uuid))


__all__ = ('Mapping', 'Mappings')



from typing import TypeVar, Callable

from lenses import UnboundLens

from amino import IO, Path, Left, do, Do
from amino.json import decode_json, dump_json

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.io.state import NS
from ribosome.nvim.io.api import N
from ribosome.config.settings import project_state_dir

A = TypeVar('A')
D = TypeVar('D')


def mkdir(dir: Path) -> IO[None]:
    return IO.delay(dir.mkdir, parents=True, exist_ok=True)


@do(NvimIO[Path])
def state_file(name: str) -> Do:
    dir = yield project_state_dir.value_or_default_fatal()
    yield N.from_io(mkdir(dir))
    yield N.pure(dir / f'{name}.json')


@do(NvimIO[A])
def load_json_data_from(name: str, file: Path) -> Do:
    exists = yield N.from_io(IO.delay(file.exists))
    if exists:
        json = yield N.from_io(IO.delay(file.read_text))
        yield N.pure(decode_json(json))
    else:
        yield N.pure(Left(f'state file {file} does not exist'))


@do(NvimIO[A])
def load_json_data(name: str) -> Do:
    file = yield state_file(name)
    yield load_json_data_from(name, file)


@do(NS[D, None])
def load_json_state(name: str, store: UnboundLens) -> Do:
    state = yield NS.lift(load_json_data(name))
    yield state.cata(lambda a: NS.pure(None), lambda d: NS.modify(store.set(d)))


@do(NvimIO[None])
def store_json_data(name: str, data: A) -> Do:
    file = yield state_file(name)
    json = yield N.from_either(dump_json(data))
    yield N.from_io(IO.delay(file.write_text, json))
    yield N.pure(None)


@do(NS[D, None])
def store_json_state(name: str, fetch: Callable[[D], A]) -> Do:
    payload = yield NS.inspect(fetch)
    yield NS.lift(store_json_data(name, payload))


__all__ = ('load_json_state', 'store_json_data', 'store_json_state')



from concurrent.futures import Future
from typing import Callable, Tuple, Generic, TypeVar

import msgpack

from amino import do, Do, IO, Right, Left
from amino.case import Case
from amino.logging import module_log

from ribosome.rpc.receive import (cons_receive, ReceiveResponse, ReceiveError, Receive, ReceiveRequest,
                                  ReceiveNotification, ReceiveExit, ReceiveUnknown)
from ribosome.rpc.comm import Comm
from ribosome.rpc.concurrency import Requests, RpcConcurrency
from ribosome.rpc.data.rpc import Rpc
from ribosome.rpc.data.rpc_type import BlockingRpc

log = module_log()
A = TypeVar('A')


def resolve_rpc(requests: Requests, id: int) -> IO[Tuple[Future, Rpc]]:
    return IO.delay(requests.to_vim.pop, id)


@do(IO[None])
def publish_response_from_vim(requests: Requests, data: ReceiveResponse) -> Do:
    fut = yield resolve_rpc(requests, data.id)
    yield IO.delay(fut.set_result, Right(data.data))


@do(IO[None])
def publish_error_from_vim(requests: Requests, err: ReceiveError) -> Do:
    fut = yield resolve_rpc(requests, err.id)
    yield IO.delay(fut.set_result, Left(err.error))


class handle_receive(Generic[A], Case[Receive, IO[None]], alg=Receive):

    def __init__(self, comm: Comm, execute_plugin_rpc: Callable[[Comm, Rpc], None]) -> None:
        self.comm = comm
        self.execute_plugin_rpc = execute_plugin_rpc

    @property
    def concurrency(self) -> RpcConcurrency:
        return self.comm.concurrency

    @property
    def requests(self) -> Requests:
        return self.concurrency.requests

    def response(self, data: ReceiveResponse) -> IO[None]:
        return self.concurrency.exclusive(publish_response_from_vim, self.requests, data)

    def error(self, err: ReceiveError) -> IO[None]:
        return self.concurrency.exclusive(publish_error_from_vim, self.requests, err)

    def request(self, receive: ReceiveRequest) -> IO[None]:
        return self.execute_plugin_rpc(self.comm, Rpc(receive.method, receive.args, BlockingRpc(receive.id)))

    def notification(self, receive: ReceiveNotification) -> IO[None]:
        return self.execute_plugin_rpc(self.comm, Rpc.nonblocking(receive.method, receive.args))

    def exit(self, receive: ReceiveExit) -> IO[None]:
        log.debug('exiting rpc session')
        return IO.pure(None)

    def unknown(self, receive: ReceiveUnknown) -> IO[None]:
        return IO.delay(log.error, f'received unknown rpc: {receive}')


def rpc_receive(comm: Comm, execute_plugin_rpc: Callable[[Comm, Rpc], None]) -> Callable[[bytes], IO[None]]:
    @do(IO[None])
    def on_read(blob: bytes) -> Do:
        data = yield IO.delay(msgpack.unpackb, blob).recover_with(lambda e: IO.failed(f'failed to unpack: {e}'))
        receive = cons_receive(data)
        yield handle_receive(comm, execute_plugin_rpc)(receive)
    return on_read


__all__ = ('rpc_receive',)



from amino.options import env_xdg_data_dir
from amino import Either, Path, Right, Try, Nil, Left, Eval, Do
from amino.do import do
from amino.boolean import true

from ribosome.nvim.io.compute import NvimIO
from ribosome.config.setting import EvalSetting, list_setting, path_setting, str_setting, bool_setting
from ribosome.nvim.io.api import N
from ribosome.nvim.api.rpc import plugin_name


@do(Either[str, str])
def project_name_from_path() -> Do:
    cwd = yield Try(Path.cwd)
    name = cwd.name
    yield Right(name) if name else Left('cwd broken')


@do(NvimIO[Either[str, Path]])
def state_dir_with_name() -> Do:
    plugin = yield plugin_name()
    base = yield state_dir.value_or_default_fatal()
    pro_name = yield proteome_main_name.value
    sess_name = yield ribosome_session_name.value
    path = pro_name.o(sess_name).o(project_name_from_path) / (lambda a: base / plugin / a)
    yield N.pure(path)


state_dir_help = '''This directory is used to persist the plugin's current state.'''
state_dir_base_default = env_xdg_data_dir.value / Path | (Path.home() / '.local' / 'share')
proteome_name_help = 'If **proteome** is installed, the session name is obtained from the main project name.'
session_name_help = 'A custom session name for the state dir can be specified.'
components_help = '''The plugin can run an arbitrary set of sub-components that inherit the class `Component`.
They receive all messages the core of the plugin processes and have access to the plugin state.
They can define nvim request handlers with the decorator `@prog`.
The entries of this setting can either be names of the builtin components or arbitrary python module paths that define
custom components.
'''

components = list_setting('components', 'names or paths of active components', components_help, True, Right(Nil))
state_dir = path_setting('ribosome_state_dir', 'state persistence directory', state_dir_help, False,
                         Right(state_dir_base_default))
proteome_main_name = str_setting('proteome_main_name', 'project name from protoeome', proteome_name_help, False)
ribosome_session_name = str_setting('ribosome_session_name', 'project name from user var', session_name_help, False)
project_state_dir = EvalSetting('project_state_dir', Eval.always(state_dir_with_name))
run_internal_init = bool_setting('run_internal_init', 'run internal initialization handler', '', True, Right(true))


__all__ = ('components', 'state_dir', 'proteome_main_name', 'ribosome_session_name', 'project_state_dir',
           'run_internal_init')



# import os
# import abc
# import pty
# import json
# import time
# import asyncio
# import subprocess
# from datetime import datetime
# from typing import Any, Callable, Generic, TypeVar

# import neovim
# from neovim.api import Nvim

# from amino import List, Either, __, env, Path, Lists, Map, do, Do, Nil, _
# from amino.lazy import lazy
# from amino.test import fixture_path, temp_dir
# from amino.test.path import base_dir, pkg_dir
# from amino.test.spec import IntegrationSpecBase as AminoIntegrationSpecBase, default_timeout
# from amino.util.string import camelcase
# from amino.json import decode_json
# from amino.json.data import JsonError

# from ribosome.logging import Logging
# from ribosome.nvim.api.data import NvimApi
# from ribosome.nvim.api.data import NativeNvimApi
# from ribosome.nvim.api.option import option_cat
# from ribosome.nvim.api.variable import variable_set_prefixed
# from ribosome.nvim.api.function import define_function, nvim_call_function, nvim_call_json
# from ribosome.nvim.io.compute import NvimIO
# from ribosome.nvim.api.command import nvim_command


# def wait_for(cond: Callable[[], bool], timeout: float=None, intval: float=0.1) -> bool:
#     to = default_timeout if timeout is None else timeout
#     start = datetime.now()
#     while (not cond() and (datetime.now() - start).total_seconds() < to):
#         time.sleep(intval)
#     return cond()


# class IntegrationSpecBase(AminoIntegrationSpecBase):

#     def setup(self) -> None:
#         AminoIntegrationSpecBase.setup(self)

#     def teardown(self) -> None:
#         AminoIntegrationSpecBase.teardown(self)


# class VimIntegrationSpecI(abc.ABC):

#     @abc.abstractproperty
#     def _log_out(self) -> List[str]:
#         ...


# class VimIntegrationSpec(VimIntegrationSpecI, IntegrationSpecBase, Logging):

#     def __init__(self) -> None:
#         IntegrationSpecBase.__init__(self)
#         self.tmux_nvim = 'RIBOSOME_TMUX_SPEC' in env
#         self.tmux_nvim_external = False
#         self.tmux_pane_id = None
#         self.keep_tmux_pane = False
#         self.vimlog = temp_dir('log') / 'vim'
#         self.nvim_cmdline = List('nvim', f'-V{self.vimlog}', '-n', '-u', 'NONE')
#         self.log_format = '{levelname} {name}:{message}'
#         self.subproc = None
#         self.ribo_tmux = None
#         self._debug = 'RIBOSOME_DEVELOPMENT' in env

#     def setup(self) -> None:
#         IntegrationSpecBase.setup(self)
#         env['RIBOSOME_SPEC'] = 1
#         env['AMINO_DEVELOPMENT'] = 1
#         env['RIBOSOME_PKG_DIR'] = str(pkg_dir())
#         self.logfile = temp_dir('log') / self.__class__.__name__
#         self.logfile.touch()
#         os.environ['RIBOSOME_LOG_FILE'] = str(self.logfile)
#         os.environ['RIBOSOME_FILE_LOG_FMT'] = self.log_format
#         self._pre_start_neovim()
#         self.start_neovim()
#         self._post_start_neovim()
#         self._pre_start()
#         self._start_plugin()
#         self._post_start()

#     def start_neovim(self) -> None:
#         if self.tmux_nvim:
#             self.start_neovim_tmux_pane()
#         elif self.tmux_nvim_external:
#             self.start_neovim_tmux_bg()
#         else:
#             self.start_neovim_embedded()
#         self.vim = self.create_nvim_api(self.neovim)

#     @property
#     def python_path(self) -> str:
#         global_path = env['PYTHONPATH'] | ''
#         return '{}:{}'.format(self.project_path, global_path)

#     @property
#     def vim_proc_env(self) -> Map:
#         return Map(
#             NVIM_LISTEN_ADDRESS=self.nvim_socket,
#             PYTHONPATH=self.python_path,
#             RIBOSOME_LOG_FILE=self.logfile,
#             RIBOSOME_FILE_LOG_FMT=f'\'{self.log_format}\'',
#             AMINO_DEVELOPMENT='1',
#         )

#     def start_neovim_embedded(self) -> None:
#         ''' start an embedded vim session that loads no init.vim.
#         **self.vimlog** is set as log file. aside from being convenient, this is crucially necessary, as the first use
#         of the session will block if stdout is used for output.
#         '''
#         argv = self.nvim_cmdline.cat('--embed')
#         self.neovim = neovim.attach('child', argv=argv)

#     @lazy
#     def nvim_socket(self) -> str:
#         return str(temp_dir('nvim_sockets') / List.random_string())

#     @property
#     def project_path(self) -> str:
#         return str(base_dir().parent)

#     def connect_tmux(self, external: bool) -> Any:
#         try:
#             from chiasma.test.tmux_spec import tmux_spec_socket
#             from chiasma.tmux import Tmux
#         except ImportError:
#             raise Exception('install chiasma to run nvim in a tmux pane')
#         else:
#             return Tmux.cons(socket=tmux_spec_socket if external else None)

#     @property
#     def tmux_window_external(self) -> Any:
#         self.ribo_tmux = self.connect_tmux(True)
#         from chiasma.io.compute import TmuxIO
#         from chiasma.commands.window import windows
#         @do(TmuxIO)
#         def run() -> Do:
#             ws = yield windows()
#             yield TmuxIO.from_maybe(ws.head, 'tmux contains no windows')
#         return run().unsafe(self.ribo_tmux)

#     @property
#     def tmux_window_local(self) -> Any:
#         session = next(s for s in server.sessions if int(s['session_attached']) >= 1)
#         return session.attached_window

#     @lazy
#     def tmux_window(self) -> Any:
#         return self.tmux_window_external if self.tmux_nvim_external else self.tmux_window_local

#     def start_neovim_tmux_bg(self) -> None:
#         conf = fixture_path('conf', 'tmux.conf')
#         args = ['tmux', '-L', self.tmux_socket, '-f', str(conf)]
#         master, slave = pty.openpty()
#         self.subproc = subprocess.Popen(args, stdout=slave, stdin=slave, stderr=slave)
#         env_args = self.vim_proc_env.map2(lambda k, v: f'{k}={v}').cons('env')
#         cmd = env_args + self.nvim_cmdline
#         from chiasma.commands.pane import send_keys
#         send_keys(0, List(cmd.join_tokens)).unsafe(self.ribo_tmux)
#         wait_for(Path(self.nvim_socket).is_socket)

#     def start_neovim_tmux_pane(self) -> None:
#         self.ribo_tmux = self.connect_tmux(False)
#         env_args = self.vim_proc_env.map2(lambda k, v: f'{k}={v}').cons('env')
#         cmd = env_args + self.nvim_cmdline
#         pid = os.getpid()
#         from chiasma.io.compute import TmuxIO
#         from chiasma.command import simple_tmux_cmd_attrs
#         from psutil import Process
#         @do(TmuxIO)
#         def run() -> Do:
#             ps = yield simple_tmux_cmd_attrs('list-panes', Nil, List('pane_pid', 'window_id'))
#             def match(pane: Map[str, str]) -> bool:
#                 return Lists.wrap(Process(int(pane['pane_pid'])).children(recursive=True)).map(_.pid).contains(pid)
#             current_pane = yield TmuxIO.from_maybe(ps.find(match), 'vim pane not found')
#             window_id = current_pane['window_id']
#             pane = yield simple_tmux_cmd_attrs('split-window', List('-t', window_id, '-d', '-P') + cmd, List('pane_id'))
#             return pane[0]['pane_id'][1:]
#         self.tmux_pane_id = run().unsafe(self.ribo_tmux)
#         wait_for(Path(self.nvim_socket).is_socket)
#         self.neovim = neovim.attach('socket', path=self.nvim_socket)
#         self.neovim.command('python3 sys.path.insert(0, \'{}\')'.format(self.python_path))

#     def _cleanup_tmux(self) -> None:
#         if self.tmux_pane_id is not None and not self.keep_tmux_pane:
#             from chiasma.commands.pane import close_pane_id
#             close_pane_id(self.tmux_pane_id).unsafe(self.ribo_tmux)

#     def create_nvim_api(self, vim: Nvim) -> NvimApi:
#         return NativeNvimApi(self.plugin_name(), vim._session)

#     @abc.abstractmethod
#     def plugin_name(self) -> str:
#         ...

#     def plugin_short_name(self) -> str:
#         return self.plugin_name()

#     def full_cmd_prefix(self) -> str:
#         return camelcase(self.plugin_name())

#     def short_cmd_prefix(self) -> str:
#         return camelcase(self.plugin_short_name())

#     def teardown(self) -> None:
#         IntegrationSpecBase.teardown(self)
#         if self._debug:
#             self._log_out.foreach(self.log.info)
#         if self.tmux_nvim:
#             self._cleanup_tmux()
#         if self.subproc is not None:
#             self.subproc.kill()
#         if self.tmux_nvim_external:
#             from chiasma.io.compute import TmuxIO
#             TmuxIO.write('kill-server').unsafe(self.ribo_tmux)

#     def _pre_start_neovim(self) -> None:
#         pass

#     def _post_start_neovim(self) -> None:
#         rtp = fixture_path('config', 'rtp')
#         option_cat('runtimepath', List(rtp)).unsafe(self.vim)
#         variable_set_prefixed('debug', True)

#     def _pre_start(self) -> None:
#         pass

#     def _start_plugin(self) -> None:
#         pass

#     def _post_start(self) -> None:
#         pass

#     @property
#     def _log_out(self) -> List[str]:
#         return List.wrap(self.logfile.read_text().splitlines())

#     def _cmd(self, cmd: str, args: List[str]) -> str:
#         return args.cons(cmd).join_tokens

#     def cmd(self, cmd: str, *args: str) -> NvimIO[None]:
#         return nvim_command(cmd, *args, verbose=True)

#     def json_cmd(self, cmd: str, *args: str, **data: Any) -> NvimIO[None]:
#         return json_cmd(cmd, *args, **data)

#     @property
#     def content(self) -> List[str]:
#         return self.vim.buffer.content

#     def program_log(self) -> Either[str, List['str']]:
#         return nvim_call_json(f'{self.full_cmd_prefix()}ProgramLog').unsafe(self.vim)

#     @property
#     def state(self) -> Any:
#         return nvim_call_json(f'{self.full_cmd_prefix()}State').unsafe(self.vim)


# D = TypeVar('D')


# class AutoPluginIntegrationSpec(Generic[D], VimIntegrationSpec):

#     def setup(self) -> None:
#         self.log_format = '{message}'
#         del env['AMINO_COMPILE_COCO']
#         super().setup()

#     def module(self) -> str:
#         return self.plugin_name()

#     @property
#     def autostart_plugin(self) -> bool:
#         return True

#     def _start_plugin(self) -> None:
#         if self.autostart_plugin:
#             self.start_plugin()

#     def start_plugin(self) -> None:
#         stderr_handler_name = 'RibosomeSpecStderr'
#         stderr_handler_body = '''
#         let err = substitute(join(a:data, '\\r'), '"', '\\"', 'g')
#         python3 import amino
#         python3 from ribosome.logging import ribosome_envvar_file_logging
#         python3 ribosome_envvar_file_logging()
#         execute 'python3 amino.amino_log.error(f"""error starting rpc job on channel ' . a:id . ':\\r' . err . '""")'
#         '''
#         define_function(stderr_handler_name, List('id', 'data', 'event'), stderr_handler_body).unsafe(self.vim)
#         cmd = f'from ribosome.host import start_module; start_module({self.module()!r})'
#         args = ['python3', '-c', cmd]
#         opts = dict(rpc=True, on_stderr=stderr_handler_name)
#         nvim_call_function('jobstart', args, opts).unsafe(self.vim)

#     def _post_start(self) -> None:
#         if self.autostart_plugin:
#             self.pvar_becomes('started', True)

#     def send_json(self, data: str) -> None:
#         self.vim.call(f'{self.plugin_name()}Send', data)


# __all__ = ('VimIntegrationSpec', 'ExternalIntegrationSpec', 'AutoPluginIntegrationSpec')



from typing import Callable

from kallikrein import Expectation

from amino import Do, do

from ribosome.nvim.io.compute import NvimIO
from ribosome.test.config import TestConfig
from ribosome.nvim.api.data import StrictNvimApi
from ribosome.test.run import run_test_io
from ribosome.nvim.io.api import N
from ribosome.test.prog import init_test_state
from ribosome.test.integration.run import StrictRequestHandler
from ribosome.nvim.io.state import NS
from ribosome.data.plugin_state import PS


def setup_strict_test_nvim(config: TestConfig) -> StrictNvimApi:
    handler = StrictRequestHandler(config.request_handler, config.function_handler, config.command_handler)
    return StrictNvimApi.cons(config.config.basic.name, vars=config.vars, request_handler=handler)


def unit_test(config: TestConfig, io: Callable[[], NS[PS, Expectation]]) -> Expectation:
    initial_nvim = setup_strict_test_nvim(config)
    @do(NvimIO[Expectation])
    def run() -> Do:
        state = yield init_test_state(config)
        yield io().run_a(state)
    return run_test_io(N.to_io_a, run(), initial_nvim)


__all__ = ('unit_test',)



from typing import Callable, Any

from kallikrein import Expectation

from amino import do, Do, IO, List, Lists
from amino.json import dump_json
from amino.logging import module_log
from amino.string.hues import green

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.io.api import N
from ribosome.nvim.api.function import define_function, nvim_call_function
from ribosome.nvim.api.rpc import nvim_quit
from ribosome.nvim.api.variable import pvar_becomes, variable_set_prefixed
from ribosome.nvim.io.data import NResult
from ribosome.test.config import TestConfig
from ribosome.test.integration.rpc import setup_test_nvim_embed
from ribosome.test.run import run_test_io

log = module_log()
stderr_handler_name = 'RibosomeSpecStderr'
stderr_handler_body = '''
let err = substitute(join(a:data, '\\r'), '"', '\\"', 'g')
python3 import amino
python3 from ribosome.logging import ribosome_envvar_file_logging
python3 ribosome_envvar_file_logging()
execute 'python3 amino.amino_log.error(f"""error starting rpc job on channel ' . a:id . ':\\r' . err . '""")'
'''


@do(NvimIO[None])
def start_plugin_embed(config: TestConfig) -> Do:
    yield define_function(stderr_handler_name, List('id', 'data', 'event'), stderr_handler_body)
    json = yield N.from_either(dump_json(config.config))
    cmd = f'from ribosome.host import start_json_config; start_json_config({json!r})'
    args = ['python3', '-c', cmd]
    opts = dict(rpc=True, on_stderr=stderr_handler_name)
    yield nvim_call_function('jobstart', args, opts)


def cleanup(config: TestConfig) -> Callable[[NResult], NvimIO[None]]:
    @do(NvimIO[None])
    def cleanup(result: NResult) -> Do:
        yield nvim_quit()
        yield nvim_quit()
        runtime_log = yield N.from_io(IO.delay(config.log_file.read_text))
        if runtime_log:
            log.info(green('plugin output:'))
            Lists.lines(runtime_log) % log.info
            log.info('')
    return cleanup


def plugin_test(config: TestConfig, io: Callable[..., NvimIO[Expectation]], *a: Any, **kw: Any) -> Expectation:
    @do(NvimIO[None])
    def run_nvim_io() -> Do:
        yield variable_set_prefixed('components', config.components)
        yield config.pre()
        yield start_plugin_embed(config)
        yield pvar_becomes('started', True)
        yield io(*a, **kw)
    @do(IO[Expectation])
    def run() -> Do:
        nvim = yield setup_test_nvim_embed(config)
        yield N.to_io_a(N.ensure(run_nvim_io(), cleanup(config)), nvim.api)
    return run_test_io(run)


__all__ = ('plugin_test',)



from typing import Callable, Tuple, Any, TypeVar, Generic

from amino import Lists, Map, List, __, Right, Nil

from kallikrein.expectable import Expectable
from amino.dat import Dat
from amino.lenses.lens import lens
from amino.logging import module_log

from ribosome.nvim.io.state import NS
from ribosome.config.config import Config
from ribosome.config.component import ComponentData
from ribosome.nvim.api.data import NvimApi, StrictNvimApi
from ribosome.nvim.io.compute import NvimIO
from ribosome.data.plugin_state import PluginState
from ribosome.nvim.api.variable import variable_set
from ribosome.nvim.io.api import N
from ribosome.compute.program import Program
from ribosome.config.resources import Resources
from ribosome.compute.output import ProgOutput
from ribosome.compute.prog import Prog
from ribosome.test.klk.expectable import kn
from ribosome.nvim.io.data import NResult
from ribosome.rpc.state import cons_state
from ribosome.components.internal.update import update_components
from ribosome.rpc.comm import StateGuard
from ribosome.rpc.to_plugin import run_programs, run_program
from ribosome.rpc.data.rpc import RpcArgs
from ribosome.test.request import Handler, StrictRequestHandler, no_handler

log = module_log()
A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')
CC = TypeVar('CC')
DIO = TypeVar('DIO')
R = TypeVar('R')
ProgInt = Callable[[ProgOutput], Callable[[R], Prog[B]]]


class RequestHelperBuilder(Generic[D, CC], Dat['RequestHelperBuilder[D, CC]']):

    def __init__(
            self,
            config: Config[D, CC],
            comps: List[str],
            vars: Map[str, Any]=Map(),
            io_interpreter: ProgInt=None,
            logger: Program[None]=None,
    ) -> None:
        self.config = config
        self.comps = comps
        self.vars = vars
        self.io_interpreter = io_interpreter
        self.logger = logger

    def create(self, cons_vim: Callable[[dict], NvimApi]) -> 'RequestHelper[D, CC]':
        comps_var = Map({f'{self.config.basic.name}_components': self.comps})
        vim = cons_vim(self.vars ** comps_var)
        state0 = cons_state(self.config, self.io_interpreter, self.logger)
        state = update_components(Right(self.comps)).run_s(state0).get_or_raise()
        return RequestHelper(vim, self.config, state)

    def nvim(self, vim: NvimApi) -> 'RequestHelper[D, CC]':
        def cons_vim(vars: dict) -> NvimApi:
            for k, v in vars.items():
                variable_set(k, v).unsafe(vim)
            return vim
        return self.create(cons_vim)

    def strict(
            self,
            request_handler: Handler=no_handler,
            function_handler: Handler=no_handler,
            command_handler: Handler=no_handler,
    ) -> 'RequestHelper[D, CC]':
        def cons_vim(vars: Map[str, Any]) -> StrictNvimApi:
            handler = StrictRequestHandler(request_handler, function_handler, command_handler)
            return StrictNvimApi(self.config.basic.name, vars, handler, Nil)
        return self.create(cons_vim)


class RequestHelper(Generic[D, CC], Dat['RequestHelper[D, CC]']):

    @staticmethod
    def nvim(config: Config[D, CC], vim: NvimApi, *comps: str, **kw: Any) -> 'RequestHelper[D, CC]':
        return RequestHelper.cons(config, *comps, **kw).nvim(vim)

    @staticmethod
    def strict(
            config: Config,
            *comps: str,
            request_handler: Handler=no_handler,
            **kw: Any,
    ) -> 'RequestHelper':
        return RequestHelper.cons(config, *comps, **kw).strict(request_handler)

    @staticmethod
    def cons(config: Config[D, CC], *comps: str, **kw: Any) -> RequestHelperBuilder[D, CC]:
        return RequestHelperBuilder(config, Lists.wrap(comps), **kw)

    def __init__(self, vim: NvimApi, config: Config, state: PluginState[D, CC]) -> None:
        self.vim = vim
        self.config = config
        self.state = state

    @property
    def guard(self) -> StateGuard[D]:
        return StateGuard.cons(self.state)

    def resolve_program(self, name: str) -> List[Program]:
        return self.guard.state.program_by_name(name)

    def run_programs(self, name: str, args: tuple=()) -> NS[PluginState[D, CC], Any]:
        programs = self.guard.state.programs_by_name(name)
        return run_programs(programs, RpcArgs.cons(Lists.wrap(args)))

    def run(self, name: str, args=(), warn_error: bool=True) -> NvimIO[Tuple[PluginState, Any]]:
        program = self.resolve_program(name).get_or_raise()
        def check_error(result: NResult) -> NvimIO[None]:
            if warn_error:
                log.error(f'NvimIO failed: {result}')
            return N.unit
        io = run_program(program, RpcArgs.cons(Lists.wrap(args))).run(self.state)
        return N.ensure_failure(io, check_error)

    def run_s(self, name: str, args=(), warn_error: bool=True) -> NvimIO[PluginState]:
        return self.run(name, args).map2(lambda s, a: s)

    def run_a(self, name: str, args=(), warn_error: bool=True) -> NvimIO[Any]:
        return self.run(name, args).map2(lambda s, a: a)

    def unsafe_run(self, name: str, args=(), warn_error: bool=True) -> Tuple[PluginState, Any]:
        return self.run(name, args=args).unsafe(self.vim)

    def unsafe_run_s(self, name: str, args=(), warn_error: bool=True) -> PluginState:
        return self.run_s(name, args=args).unsafe(self.vim)

    def unsafe_run_a(self, name: str, args=(), warn_error: bool=True) -> Any:
        return self.run_a(name, args=args).unsafe(self.vim)

    def update_data(self, **kw: Any) -> 'RequestHelper[D, CC]':
        return lens.state.data.modify(__.copy(**kw))(self)

    def update_component(self, name: str, **kw: Any) -> 'RequestHelper[D, CC]':
        return self.mod.state(__.modify_component_data(name, __.copy(**kw)))

    def component_res(self, data: C) -> Resources[C, CC]:
        return self.state.resources_with(ComponentData(self.state.data, data))

    def component_res_for(self, name: str) -> Resources[C, CC]:
        data = self.state.data_by_name(name).get_or_raise()
        return self.state.resources_with(ComponentData(self.state.data, data))

    def k(self, f: Callable[..., NvimIO[A]], *a: Any, **kw: Any) -> Expectable:
        return kn(self.vim, f, *a, **kw)

    def k_s(self, prog: str, *a: Any, **kw: Any) -> Expectable:
        return self.k(self.run_s, prog, a, **kw)


def request_helper(
        config: Config[D, CC],
        *comps: str,
        vars: Map[str, Any]=Map(),
        io_interpreter: ProgInt=None,
) -> NvimIO[RequestHelper[D, CC]]:
    return N.delay(lambda v: RequestHelper.nvim(config, v, *comps, vars=vars, io_interpreter=io_interpreter))


__all__ = ('RequestHelper', 'request_helper')
